## Run Ignite VMs declaratively

Flags can be convenient for simple use cases, but have many limitations.
In more advanced use-cases, and to eventually allow GitOps flows, there is
an other way: telling Ignite what to do _declaratively_, using a file containing
an API object.

The first commands to support this feature is `ignite run` and `ignite create`.
An example file as follows: 

```yaml
apiVersion: ignite.weave.works/v1alpha1
kind: VM
metadata:
  name: test-vm
spec:
  image:
    ociClaim:
      ref: weaveworks/ignite-ubuntu
  cpus: 2
  diskSize: 3GB
  memory: 800MB
```

This API object specifies a need for 2 vCPUs, 800MB of RAM and 3GB of disk.

We can tell Ignite to make this happen using simply:

```console
$ ignite run --config test-vm.yaml
```

The full reference format for the `VM` kind is as follows:

```yaml
apiVersion: ignite.weave.works/v1alpha1
kind: VM
metadata:
  # Automatically set when the object is created
  created: [RFC3339-formatted date]
  # Mandatory, the name of the VM
  name: [string]
  # Optional, autogenerated if not specified
  uid: [16-char hex UID]
  # Optional, a string-string map with label keys and values
  labels:
    foo: bar
  # Optional, a string-string map with annotation keys and values
  annotations:
    foo: bar
spec:
  # Optional, how many vCPUs that should be allocated for the VM
  # Default: 1
  cpus: [uint64]
  # Optional, how much RAM that should be allocated for the VM
  # Default: 512MB
  memory: [size]
  # Optional, how much free writable space the VM should have at runtime
  # Default: 4GB
  diskSize: [size]

  image:
    ociClaim:
      # Required: What OCI image to use as the VM's rootfs
      # For example: weaveworks/ignite-ubuntu:latest
      ref: [OCI image reference]
  kernel:
    # Optional, the kernel commandline to be specified
    # Default: "console=ttyS0 reboot=k panic=1 pci=off ip=dhcp"
    cmdLine: [string]
    ociClaim:
      # Required: What OCI image to get the kernel executable, and optionally modules from
      # Default: weaveworks/ignite-kernel:4.19.47
      ref:  [OCI image reference]
  
  network:
    # Optional: What networking mode to use. Available options are: "docker-bridge" and "cni"
    # Default: docker-bridge
    mode: [networking-mode]
    # Optional, an array of portmappings that map ports bound to the VM to the host
    # Default: unset, no portmappings
    ports:
    # This example maps 0.0.0.0:6443 inside of the VM to 0.0.0.0:443 on the physical host
    - hostPort: 433
      vmPort: 6443
  
  # Optional, an array of files to copy into the VM at runtime
  # Default: unset, no filemappings
  copyFiles:
  # This example copies a Kubernetes KubeConfig file from /etc/kubernetes/admin.conf
  # on the host to /home/lucas/.kube/config inside of the VM
  - hostPath: /etc/kubernetes/admin.conf
    vmPath: /home/lucas/.kube/config
  
  # Optional, provides automation to easily access your VM with the "ignite ssh" command
  # If "ssh: true" is set, Ignite will generate an SSH key and copy the public key into the VM
  # This allows for automatic "ignite ssh" logins.
  # Alternatively: specify a path to a public key to put in authorized_keys in the VM.
  # Default: unset, no actions regarding SSH automation
  ssh: [true, or public key path]
```

You can find the full API reference in the [api/](../api) subfolder of the project.
